/* -*-c-*-
 *
 * Definitions for cipher block chaining mode
 *
 * (c) 1999 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

#ifndef CATACOMB_CBC_DEF_H
#define CATACOMB_CBC_DEF_H

#ifdef __cplusplus
  extern "C" {
#endif

/*----- Header files ------------------------------------------------------*/

#include <string.h>

#include <mLib/bits.h>
#include <mLib/sub.h>

#ifndef CATACOMB_ARENA_H
#  include "arena.h"
#endif

#ifndef CATACOMB_BLKC_H
#  include "blkc.h"
#endif

#ifndef CATACOMB_GCIPHER_H
#  include "gcipher.h"
#endif

#ifndef CATACOMB_PARANOIA_H
#  include "paranoia.h"
#endif

/*----- Macros ------------------------------------------------------------*/

/* --- @CBC_DEF@ --- *
 *
 * Arguments:	@PRE@, @pre@ = prefixes for the underlying block cipher
 *
 * Use:		Creates an implementation for CBC stealing mode.
 */

#define CBC_DEF(PRE, pre) CBC_DEFX(PRE, pre, #pre, #pre)

#define CBC_DEFX(PRE, pre, name, fname)					\
									\
/* --- @pre_cbcgetiv@ --- *						\
 *									\
 * Arguments:	@const pre_cbcctx *ctx@ = pointer to CBC context block	\
 *		@void *iv@ = pointer to output data block		\
 *									\
 * Returns:	---							\
 *									\
 * Use:		Reads the currently set IV.  Reading and setting an IV	\
 *		is transparent to the CBC encryption or decryption	\
 *		process.						\
 */									\
									\
void pre##_cbcgetiv(const pre##_cbcctx *ctx, void *iv)			\
  { BLKC_STORE(PRE, iv, ctx->a); }					\
									\
/* --- @pre_cbcsetiv@ --- *						\
 *									\
 * Arguments:	@pre_cbcctx *ctx@ = pointer to CBC context block	\
 *		@cnost void *iv@ = pointer to IV to set			\
 *									\
 * Returns:	---							\
 *									\
 * Use:		Sets the IV to use for subsequent encryption.		\
 */									\
									\
void pre##_cbcsetiv(pre##_cbcctx *ctx, const void *iv)			\
  { BLKC_LOAD(PRE, ctx->a, iv); }					\
									\
/* --- @pre_cbcsetkey@ --- *						\
 *									\
 * Arguments:	@pre_cbcctx *ctx@ = pointer to CBC context block	\
 *		@const pre_ctx *k@ = pointer to cipher context		\
 *									\
 * Returns:	---							\
 *									\
 * Use:		Sets the CBC context to use a different cipher key.	\
 */									\
									\
void pre##_cbcsetkey(pre##_cbcctx *ctx, const pre##_ctx *k)		\
  { ctx->ctx = *k; }							\
									\
/* --- @pre_cbcinit@ --- *						\
 *									\
 * Arguments:	@pre_cbcctx *ctx@ = pointer to cipher context		\
 *		@const void *key@ = pointer to the key buffer		\
 *		@size_t sz@ = size of the key				\
 *		@const void *iv@ = pointer to initialization vector	\
 *									\
 * Returns:	---							\
 *									\
 * Use:		Initializes a CBC context ready for use.  The @iv@	\
 *		argument may be passed as a null pointer to set a zero	\
 *		IV.  Apart from that, this call is equivalent to calls	\
 *		to @pre_init@, @pre_cbcsetkey@ and @pre_cbcsetiv@.	\
 */									\
									\
void pre##_cbcinit(pre##_cbcctx *ctx,					\
		   const void *key, size_t sz,				\
		   const void *iv)					\
{									\
  static const octet zero[PRE##_BLKSZ] = { 0 };				\
									\
  pre##_init(&ctx->ctx, key, sz);					\
  BLKC_LOAD(PRE, ctx->a, iv ? iv : zero);				\
}									\
									\
/* --- @pre_cbcencrypt@ --- *						\
 *									\
 * Arguments:	@pre_cbcctx *ctx@ = pointer to CBC context block	\
 *		@const void *src@ = pointer to source data		\
 *		@void *dest@ = pointer to destination data		\
 *		@size_t sz@ = size of block to be encrypted		\
 *									\
 * Returns:	---							\
 *									\
 * Use:		Encrypts a block with a block cipher in CBC mode, with	\
 *		ciphertext stealing and other clever tricks.		\
 *		Essentially, data can be encrypted in arbitrary sized	\
 *		chunks, although decryption must use the same chunks.	\
 */									\
									\
void pre##_cbcencrypt(pre##_cbcctx *ctx,				\
		      const void *src, void *dest,			\
		      size_t sz)					\
{									\
  const octet *s = src;							\
  octet *d = dest;							\
  octet b[PRE##_BLKSZ], bb[PRE##_BLKSZ];				\
  octet y;								\
  unsigned i;								\
									\
  /* --- Empty blocks are trivial --- */				\
									\
  if (!sz) return;							\
									\
  /* --- Extra magical case for a short block --- *			\
   *									\
   * Encrypt the IV, then exclusive-or the plaintext with the octets	\
   * of the encrypted IV, shifting ciphertext octets in instead.  This	\
   * basically switches over to CFB.					\
   */									\
									\
  if (sz < PRE##_BLKSZ) {						\
    pre##_eblk(&ctx->ctx, ctx->a, ctx->a);				\
    BLKC_STORE(PRE, b, ctx->a);						\
    if (!d) d = bb;							\
    for (i = 0; i < sz; i++) d[i] = b[i] ^ (s ? s[i] : 0);		\
    memmove(b, b + sz, PRE##_BLKSZ - sz);				\
    memcpy(b + PRE##_BLKSZ - sz, d, sz);				\
    BLKC_LOAD(PRE, ctx->a, b);						\
    return;								\
  }									\
									\
  /* --- Do the main chunk of encryption --- *				\
   *									\
   * This will do the whole lot if it's a whole number of blocks.  For	\
   * each block, XOR it with the previous ciphertext in @iv@, encrypt,	\
   * and keep a copy of the ciphertext for the next block.		\
   */									\
									\
  while (sz >= 2*PRE##_BLKSZ || sz == PRE##_BLKSZ) {			\
    if (s) { BLKC_XLOAD(PRE, ctx->a, s); s += PRE##_BLKSZ; }		\
    pre##_eblk(&ctx->ctx, ctx->a, ctx->a);				\
    if (d) { BLKC_STORE(PRE, d, ctx->a); d += PRE##_BLKSZ; }		\
    sz -= PRE##_BLKSZ;							\
  }									\
									\
  /* --- Do the tail-end block and bit-left-over --- *			\
   *									\
   * This isn't very efficient.  That shouldn't matter much.		\
   */									\
									\
  if (sz) {								\
									\
    /* --- Let @sz@ be the size of the partial block --- */		\
									\
    sz -= PRE##_BLKSZ;							\
									\
    /* --- First stage --- *						\
     *									\
     * XOR the complete block with the current IV, and encrypt it.  The	\
     * first part of the result is the partial ciphertext block.  Don't	\
     * write that out yet, because I've not read the partial plaintext	\
     * block.								\
     */									\
									\
    if (s) BLKC_XLOAD(PRE, ctx->a, s);					\
    pre##_eblk(&ctx->ctx, ctx->a, ctx->a);				\
    BLKC_STORE(PRE, b, ctx->a);						\
									\
    /* --- Second stage --- *						\
     *									\
     * Now XOR in the partial plaintext block, writing out the		\
     * ciphertext as I go.  Then encrypt, and write the complete	\
     * ciphertext block.						\
     */									\
									\
    if (s) s += PRE##_BLKSZ;						\
    if (d) d += PRE##_BLKSZ;						\
    for (i = 0; i < sz; i++) {						\
      y = b[i];								\
      if (s) b[i] ^= s[i];						\
      if (d) d[i] = y;							\
    }									\
    BLKC_LOAD(PRE, ctx->a, b);						\
    pre##_eblk(&ctx->ctx, ctx->a, ctx->a);				\
    if (d) BLKC_STORE(PRE, d - PRE##_BLKSZ, ctx->a);			\
  }									\
									\
  /* --- Done --- */							\
									\
  return;								\
}									\
									\
/* --- @pre_cbcdecrypt@ --- *						\
 *									\
 * Arguments:	@pre_cbcctx *ctx@ = pointer to CBC context block	\
 *		@const void *src@ = pointer to source data		\
 *		@void *dest@ = pointer to destination data		\
 *		@size_t sz@ = size of block to be encrypted		\
 *									\
 * Returns:	---							\
 *									\
 * Use:		Decrypts a block with a block cipher in CBC mode, with	\
 *		ciphertext stealing and other clever tricks.		\
 *		Essentially, data can be encrypted in arbitrary sized	\
 *		chunks, although decryption must use the same chunks.	\
 */									\
									\
void pre##_cbcdecrypt(pre##_cbcctx *ctx,				\
		      const void *src, void *dest,			\
		      size_t sz)					\
{									\
  const octet *s = src;							\
  octet *d = dest;							\
  uint32 t[PRE##_BLKSZ/4], u[PRE##_BLKSZ/4];				\
  octet b[PRE##_BLKSZ], c[PRE##_BLKSZ];					\
  octet y;								\
  unsigned i;								\
									\
  /* --- Empty blocks are trivial --- */				\
									\
  if (!sz) return;							\
									\
  /* --- Extra magical case for a short block --- *			\
   *									\
   * Encrypt the IV, then exclusive-or the ciphertext with the octets	\
   * of the encrypted IV, shifting ciphertext octets in instead.  This	\
   * basically switches over to CFB.					\
   */									\
									\
  if (sz < PRE##_BLKSZ) {						\
    pre##_eblk(&ctx->ctx, ctx->a, ctx->a);				\
    BLKC_STORE(PRE, b, ctx->a);						\
    for (i = 0; i < sz; i++) { y = s[i]; d[i] = b[i] ^ y; c[i] = y; }	\
    memmove(b, b + sz, PRE##_BLKSZ - sz);				\
    memcpy(b + PRE##_BLKSZ - sz, c, sz);				\
    BLKC_LOAD(PRE, ctx->a, b);						\
    return;								\
  }									\
									\
  /* --- Do the main chunk of decryption --- *				\
   *									\
   * This will do the whole lot if it's a whole number of blocks.  For	\
   * each block, decrypt, XOR it with the previous ciphertext in @iv@,	\
   * and keep a copy of the ciphertext for the next block.		\
   */									\
									\
  while (sz >= 2*PRE##_BLKSZ || sz == PRE##_BLKSZ) {			\
    BLKC_LOAD(PRE, t, s); s += PRE##_BLKSZ;				\
    pre##_dblk(&ctx->ctx, t, u);					\
    BLKC_XSTORE(PRE, d, u, ctx->a); d += PRE##_BLKSZ;			\
    BLKC_MOVE(PRE, ctx->a, t);						\
    sz -= PRE##_BLKSZ;							\
  }									\
									\
  /* --- Do the tail-end block and bit-left-over --- *			\
   *									\
   * This isn't very efficient.  That shouldn't matter much.		\
   */									\
									\
  if (sz) {								\
									\
    /* --- Let @sz@ be the size of the partial block --- */		\
									\
    sz -= PRE##_BLKSZ;							\
									\
    /* --- First stage --- *						\
     *									\
     * Take the complete ciphertext block, and decrypt it.  This block	\
     * is carried over for the next encryption operation.		\
     */									\
									\
    BLKC_LOAD(PRE, t, s);						\
    pre##_dblk(&ctx->ctx, t, u);					\
									\
    /* --- Second stage --- *						\
     *									\
     * XORing the first few bytes of this with the partial ciphertext	\
     * block recovers the partial plaintext block.  At the same time,	\
     * write the partial ciphertext block's contents in ready for stage	\
     * three.								\
     */									\
									\
    BLKC_STORE(PRE, b, u);						\
    s += PRE##_BLKSZ;							\
    d += PRE##_BLKSZ;							\
    for (i = 0; i < sz; i++) { y = s[i]; d[i] = b[i] ^ y; b[i] = y; }	\
									\
    /* --- Third stage --- *						\
     *									\
     * Decrypt the block we've got left, and XOR with the initial IV to	\
     * recover the complete plaintext block.				\
     */									\
									\
    BLKC_LOAD(PRE, u, b);						\
    pre##_dblk(&ctx->ctx, u, u);					\
    BLKC_XSTORE(PRE, d - PRE##_BLKSZ, u, ctx->a);			\
    BLKC_MOVE(PRE, ctx->a, t);						\
  }									\
									\
  /* --- Done --- */							\
									\
  return;								\
}									\
									\
/* --- Generic cipher interface --- */					\
									\
static const gcipher_ops gops;						\
									\
typedef struct gctx {							\
  gcipher c;								\
  pre##_cbcctx k;							\
} gctx;									\
									\
static gcipher *ginit(const void *k, size_t sz)				\
{									\
  gctx *g = S_CREATE(gctx);						\
  g->c.ops = &gops;							\
  pre##_cbcinit(&g->k, k, sz, 0);					\
  return (&g->c);							\
}									\
									\
static void gencrypt(gcipher *c, const void *s, void *t, size_t sz)	\
  { gctx *g = (gctx *)c; pre##_cbcencrypt(&g->k, s, t, sz); }		\
									\
static void gdecrypt(gcipher *c, const void *s, void *t, size_t sz)	\
  { gctx *g = (gctx *)c; pre##_cbcdecrypt(&g->k, s, t, sz); }		\
									\
static void gdestroy(gcipher *c)					\
  { gctx *g = (gctx *)c; BURN(*g); S_DESTROY(g); }			\
									\
static void gsetiv(gcipher *c, const void *iv)				\
  { gctx *g = (gctx *)c; pre##_cbcsetiv(&g->k, iv); }			\
									\
static const gcipher_ops gops = {					\
  &pre##_cbc,								\
  gencrypt, gdecrypt, gdestroy, gsetiv, 0				\
};									\
									\
const gccipher pre##_cbc = {						\
  name "-cbc", pre##_keysz, PRE##_BLKSZ,				\
  ginit									\
};									\
									\
CBC_TESTX(PRE, pre, name, fname)

/*----- Test rig ----------------------------------------------------------*/

#define CBC_TEST(PRE, pre) CBC_TESTX(PRE, pre, #pre, #pre)

#ifdef TEST_RIG

#include "modes-test.h"

/* --- @CBC_TEST@ --- *
 *
 * Arguments:	@PRE@, @pre@ = prefixes for block cipher definitions
 *
 * Use:		Standard test rig for CBC functions.
 */

#define CBC_TESTX(PRE, pre, name, fname)				\
									\
static pre##_ctx key;							\
static pre##_cbcctx ctx;						\
									\
static void pre##_cbc_test_setup(const octet *k, size_t ksz)		\
  { pre##_init(&key, k, ksz); pre##_cbcsetkey(&ctx, &key); }		\
									\
static void pre##_cbc_test_reset(const octet *iv)			\
  { pre##_cbcsetiv(&ctx, iv); }						\
									\
static void pre##_cbc_test_enc(const octet *s, octet *d, size_t sz)	\
  { pre##_cbcencrypt(&ctx, s, d, sz); }					\
									\
static void pre##_cbc_test_dec(const octet *s, octet *d, size_t sz)	\
  { pre##_cbcdecrypt(&ctx, s, d, sz); }					\
									\
int main(int argc, char *argv[])					\
{									\
  return test_encmode(fname "-cbc", PRE##_KEYSZ, PRE##_BLKSZ,		\
		      1, TEMF_REFALIGN,					\
		      pre##_cbc_test_setup, pre##_cbc_test_reset,	\
		      pre##_cbc_test_enc, pre##_cbc_test_dec,		\
		      argc, argv);					\
}

#else
#  define CBC_TESTX(PRE, pre, name, fname)
#endif

/*----- That's all, folks -------------------------------------------------*/

#ifdef __cplusplus
  }
#endif

#endif
