/* -*-c-*-
 *
 * The HAS160 message digest function
 *
 * (c) 2004 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Notes on the HAS160 hash function ---------------------------------*
 *
 * HAS160 was designed by Chae Hoon Lim and the Korean Information Security
 * Agency (KISA).  It's recommended for use with KCDSA (though I think I'm
 * happer with RIPEMD-160 or SHA1).  It's here so I can check KCDSA test
 * vectors.
 */

#ifndef CATACOMB_HAS160_H
#define CATACOMB_HAS160_H

#ifdef __cplusplus
  extern "C" {
#endif

/*----- Header files ------------------------------------------------------*/

#include <mLib/bits.h>

#ifndef CATACOMB_GHASH_H
#  include "ghash.h"
#endif

/*----- Magic numbers -----------------------------------------------------*/

#define HAS160_BUFSZ 64
#define HAS160_HASHSZ 20
#define HAS160_STATESZ 20

/*----- Data structures ---------------------------------------------------*/

typedef struct has160_ctx {
  uint32 a, b, c, d, e;			/* Chaining variables */
  uint32 nl, nh;			/* Byte count so far */
  unsigned off;				/* Offset into buffer */
  octet buf[HAS160_BUFSZ];		/* Accumulation buffer */
} has160_ctx;

/*----- Functions provided ------------------------------------------------*/

/* --- @has160_compress@ --- *
 *
 * Arguments:	@has160_ctx *ctx@ = pointer to context block
 *		@const void *sbuf@ = pointer to buffer of appropriate size
 *
 * Returns:	---
 *
 * Use:		HAS160 compression function.
 */

extern void has160_compress(has160_ctx */*ctx*/, const void */*sbuf*/);

/* --- @has160_init@ --- *
 *
 * Arguments:	@has160_ctx *ctx@ = pointer to context block to initialize
 *
 * Returns:	---
 *
 * Use:		Initializes a context block ready for hashing.
 */

extern void has160_init(has160_ctx */*ctx*/);

/* --- @has160_set@ --- *
 *
 * Arguments:	@has160_ctx *ctx@ = pointer to context block
 *		@const void *buf@ = pointer to state buffer
 *		@unsigned long count@ = current count of bytes processed
 *
 * Returns:	---
 *
 * Use:		Initializes a context block from a given state.  This is
 *		useful in cases where the initial hash state is meant to be
 *		secret, e.g., for NMAC and HMAC support.
 */

extern void has160_set(has160_ctx */*ctx*/, const void */*buf*/,
		       unsigned long /*count*/);

/* --- @has160_hash@ --- *
 *
 * Arguments:	@has160_ctx *ctx@ = pointer to context block
 *		@const void *buf@ = buffer of data to hash
 *		@size_t sz@ = size of buffer to hash
 *
 * Returns:	---
 *
 * Use:		Hashes a buffer of data.  The buffer may be of any size and
 *		alignment.
 */

extern void has160_hash(has160_ctx */*ctx*/,
			const void */*buf*/, size_t /*sz*/);

/* --- @has160_done@ --- *
 *
 * Arguments:	@has160_ctx *ctx@ = pointer to context block
 *		@void *hash@ = pointer to output buffer
 *
 * Returns:	---
 *
 * Use:		Returns the hash of the data read so far.
 */

extern void has160_done(has160_ctx */*ctx*/, void */*hash*/);

/* --- @has160_state@ --- *
 *
 * Arguments:	@has160_ctx *ctx@ = pointer to context
 *		@void *state@ = pointer to buffer for current state
 *
 * Returns:	Number of bytes written to the hash function so far.
 *
 * Use:		Returns the current state of the hash function such that
 *		it can be passed to @has160_set@.
 */

extern unsigned long has160_state(has160_ctx */*ctx*/, void */*state*/);

/*----- Generic hash interface --------------------------------------------*/

extern const gchash has160;

/*----- That's all, folks -------------------------------------------------*/

#ifdef __cplusplus
  }
#endif

#endif
