/* -*-c-*-
 *
 * Definitions for the MGF-1 mask generator
 *
 * (c) 2000 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

#ifndef CATACOMB_MGF_DEF_H
#define CATACOMB_MGF_DEF_H

#ifdef __cplusplus
  extern "C" {
#endif

/*----- Header files ------------------------------------------------------*/

#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#include <mLib/bits.h>
#include <mLib/sub.h>

#ifndef CATACOMB_ARENA_H
#  include "arena.h"
#endif

#ifndef CATACOMB_GCIPHER_H
#  include "gcipher.h"
#endif

#ifndef CATACOMB_GRAND_H
#  include "grand.h"
#endif

#ifndef CATACOMB_PARANOIA_H
#  include "paranoia.h"
#endif

/*----- Macros ------------------------------------------------------------*/

#define MGF_DEF(PRE, pre) MGF_DEFX(PRE, pre, #pre, #pre)

#define MGF_DEFX(PRE, pre, name, fname)					\
									\
/* --- Useful constants --- */						\
									\
const octet pre##_mgfkeysz[] = { KSZ_ANY, PRE##_HASHSZ };		\
									\
/* --- @pre_mgfkeybegin@, @pre_mgfkeyadd@ --- *				\
 *									\
 * Arguments:	@pre_mgfctx *k@ = pointer to context to initialize	\
 *		@const void *p@ = pointer to data to contribute		\
 *									\
 * Returns:	---							\
 *									\
 * Use:		A multi-step keying procedure for initializing an MGF	\
 *		context.  The data is contributed to a hashing context	\
 *		which is then used for mask generation.  If you only	\
 *		have a fixed buffer, you can save a lot of effort by	\
 *		simply calling @pre_mgfinit@.				\
 */									\
									\
void pre##_mgfkeybegin(pre##_mgfctx *k)					\
  { k->c = 0; k->off = 0; pre##_init(&k->k); }				\
									\
void pre##_mgfkeyadd(pre##_mgfctx *k, const void *p, size_t sz)		\
  { pre##_hash(&k->k, p, sz); }						\
									\
/* ---- @pre_mgfinit@ --- *						\
 *									\
 * Arguments:	@pre_mgfctx *k@ = pointer to context to initialize	\
 *		@const void *p@ = pointer to data to contribute		\
 *		@size_t sz@ = size of data to contribute		\
 *									\
 * Returns:	---							\
 *									\
 * Use:		A simpler interface to initialization if all of your	\
 *		keying material is in one place.			\
 */									\
									\
void pre##_mgfinit(pre##_mgfctx *k, const void *p, size_t sz)		\
  { k->c = 0; k->off = 0; pre##_init(&k->k); pre##_hash(&k->k, p, sz); } \
									\
/* --- @pre_mgfencrypt@ --- *						\
 *									\
 * Arguments:	@pre_mgfctx *k@ = pointer to masking context		\
 *		@const void *s@ = pointer to source buffer		\
 *		@void *d@ = pointer to destination buffer		\
 *		@size_t sz@ = size of buffers				\
 *									\
 * Returns:	---							\
 *									\
 * Use:		Outputs pseudorandom data, or masks an input buffer.	\
 *									\
 *		If @s@ is nonzero, the source material is exclusive-	\
 *		orred with the generated mask.  If @d@ is zero, the	\
 *		generator is simply spun around for a while, which	\
 *		isn't very useful.					\
 */									\
									\
void pre##_mgfencrypt(pre##_mgfctx *k, const void *s,			\
		      void *d, size_t sz)				\
{									\
  pre##_ctx h;								\
  const octet *ss = s;							\
  octet *dd = d;							\
  const octet *p;							\
  size_t off;								\
									\
  /* --- Empty the buffer if there's anything there --- */		\
									\
  if (k->off) {								\
    p = k->b + PRE##_HASHSZ - k->off;					\
    off = sz > k->off ? k->off : sz;					\
    sz -= off; k->off -= off;						\
    if (!dd) /* do nothing */;						\
    else if (!ss) { memcpy(dd, p, off); dd += off; }			\
    else while (off--) *dd++ = *ss++ ^ *p++;				\
  }									\
									\
  /* --- While necessary, generate some more mask --- */		\
									\
  while (sz) {								\
    STORE32(k->b, k->c); k->c++;					\
    h = k->k; pre##_hash(&h, k->b, 4); pre##_done(&h, k->b);		\
    off = sz > PRE##_HASHSZ ? PRE##_HASHSZ : sz;			\
    k->off = PRE##_HASHSZ - off;					\
    sz -= off;								\
    p = k->b;								\
    if (!dd) /* do nothing */;						\
    else if (!ss) { memcpy(dd, p, off); dd += off; }			\
    else while (off--) *dd++ = *ss++ ^ *p++;				\
  }									\
}									\
									\
/* --- @pre_mgfsetindex@ --- *						\
 *									\
 * Arguments:	@pre_mgfctx *k@ = pointer to masking context		\
 *		@uint32 *c@ = new index to set				\
 *									\
 * Returns:	---							\
 *									\
 * Use:		Sets a new index.  This may be used to step around the	\
 *		output stream in a rather crude way.			\
 */									\
									\
void pre##_mgfsetindex(pre##_mgfctx *k, uint32 c)			\
  { k->c = c; k->off = 0; }						\
									\
/* --- Generic cipher interface --- */					\
									\
static const gcipher_ops gops;						\
									\
typedef struct gctx {							\
  gcipher c;								\
  pre##_mgfctx k;							\
} gctx;									\
									\
static gcipher *ginit(const void *k, size_t sz)				\
{									\
  gctx *g = S_CREATE(gctx);						\
  g->c.ops = &gops;							\
  pre##_mgfinit(&g->k, k, sz);						\
  return (&g->c);							\
}									\
									\
static void gencrypt(gcipher *c, const void *s, void *t, size_t sz)	\
  { gctx *g = (gctx *)c; pre##_mgfencrypt(&g->k, s, t, sz); }		\
									\
static void gdestroy(gcipher *c)					\
  { gctx *g = (gctx *)c; BURN(*g); S_DESTROY(g); }			\
									\
static const gcipher_ops gops = {					\
  &pre##_mgf,								\
  gencrypt, gencrypt, gdestroy, 0, 0					\
};									\
									\
const gccipher pre##_mgf = {						\
  name "-mgf", pre##_mgfkeysz, 0,					\
  ginit									\
};									\
									\
/* --- Generic random number generator interface --- */			\
									\
typedef struct grctx {							\
  grand r;								\
  pre##_mgfctx k;							\
} grctx;								\
									\
static void grdestroy(grand *r)						\
{									\
  grctx *g = (grctx *)r;						\
  BURN(*g);								\
  S_DESTROY(g);								\
}									\
									\
static int grmisc(grand *r, unsigned op, ...)				\
{									\
  grctx *g = (grctx *)r;						\
  va_list ap;								\
  int rc = 0;								\
  va_start(ap, op);							\
									\
  switch (op) {								\
    case GRAND_CHECK:							\
      switch (va_arg(ap, unsigned)) {					\
	case GRAND_CHECK:						\
	case GRAND_SEEDINT:						\
	case GRAND_SEEDUINT32:						\
	case GRAND_SEEDBLOCK:						\
	case GRAND_SEEDRAND:						\
	  rc = 1;							\
	  break;							\
	default:							\
	  rc = 0;							\
	  break;							\
      }									\
      break;								\
    case GRAND_SEEDINT:							\
      pre##_mgfsetindex(&g->k, va_arg(ap, unsigned));			\
      break;								\
    case GRAND_SEEDUINT32:						\
      pre##_mgfsetindex(&g->k, va_arg(ap, uint32));			\
      break;								\
    case GRAND_SEEDBLOCK: {						\
      const void *p = va_arg(ap, const void *);				\
      size_t sz = va_arg(ap, size_t);					\
      pre##_hash(&g->k.k, p, sz);					\
    } break;								\
    case GRAND_SEEDRAND: {						\
      octet buf[PRE##_BUFSZ];						\
      grand *rr = va_arg(ap, grand *);					\
      rr->ops->fill(rr, buf, sizeof(buf));				\
      pre##_hash(&g->k.k, buf, sizeof(buf));				\
    } break;								\
    default:								\
      GRAND_BADOP;							\
      break;								\
  }									\
									\
  va_end(ap);								\
  return (rc);								\
}									\
									\
static octet grbyte(grand *r)						\
{									\
  grctx *g = (grctx *)r;						\
  octet o;								\
  pre##_mgfencrypt(&g->k, 0, &o, 1);					\
  return (o);								\
}									\
									\
static uint32 grword(grand *r)						\
{									\
  grctx *g = (grctx *)r;						\
  octet b[4];								\
  pre##_mgfencrypt(&g->k, 0, b, sizeof(b));				\
  return (LOAD32(b));							\
}									\
									\
static void grfill(grand *r, void *p, size_t sz)			\
  { grctx *g = (grctx *)r; pre##_mgfencrypt(&g->k, 0, p, sz); }		\
									\
static const grand_ops grops = {					\
  name "-mgf",								\
  GRAND_CRYPTO, 0,							\
  grmisc, grdestroy,							\
  grword, grbyte, grword, grand_defaultrange, grfill			\
};									\
									\
/* --- @pre_mgfrand@ --- *						\
 *									\
 * Arguments:	@const void *k@ = pointer to key material		\
 *		@size_t sz@ = size of key material			\
 *									\
 * Returns:	Pointer to a generic random number generator instance.	\
 *									\
 * Use:		Creates a random number interface wrapper around an	\
 *		MGF-1-mode hash function.				\
 */									\
									\
extern grand *pre##_mgfrand(const void *k, size_t sz)			\
{									\
  grctx *g = S_CREATE(grctx);						\
  g->r.ops = &grops;							\
  pre##_mgfinit(&g->k, k, sz);						\
  return (&g->r);							\
}									\
									\
MGF_TESTX(PRE, pre, name, fname)

/*----- Test rig ----------------------------------------------------------*/

#define MGF_TEST(PRE, pre) MGF_TESTX(PRE, pre, #pre, #pre)

#ifdef TEST_RIG

#include "modes-test.h"

/* --- @MGF_TEST@ --- *
 *
 * Arguments:	@PRE@, @pre@ = prefixes for block cipher definitions
 *
 * Use:		Standard test rig for MGF functions.
 */

#define MGF_TESTX(PRE, pre, name, fname)				\
									\
static pre##_mgfctx ctx;						\
									\
static void pre##_mgf_test_setup(const octet *k, size_t ksz)		\
  { pre##_mgfinit(&ctx, k, ksz); }					\
									\
static void pre##_mgf_test_reset(const octet *iv)			\
  { pre##_mgfsetindex(&ctx, 0); }					\
									\
static void pre##_mgf_test_enc(const octet *s, octet *d, size_t sz)	\
  { pre##_mgfencrypt(&ctx, s, d, sz); }					\
									\
int main(int argc, char *argv[])					\
{									\
  return test_encmode(fname "-mgf", 0, PRE##_HASHSZ, 1, 0,		\
		      pre##_mgf_test_setup, pre##_mgf_test_reset,	\
		      pre##_mgf_test_enc, pre##_mgf_test_enc,		\
		      argc, argv);					\
}

#else
#  define MGF_TESTX(PRE, pre, name, fname)
#endif

/*----- That's all, folks -------------------------------------------------*/

#ifdef __cplusplus
  }
#endif

#endif
