/* -*-c-*-
 *
 * Montgomery reduction
 *
 * (c) 1999 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

#ifndef CATACOMB_MPMONT_H
#define CATACOMB_MPMONT_H

#ifdef __cplusplus
  extern "C" {
#endif

/*----- Header files ------------------------------------------------------*/

#ifndef CATACOMB_MP_H
#  include "mp.h"
#endif

/*----- Notes on Montgomery reduction -------------------------------------*
 *
 * Given a little bit of precomputation, Montgomery reduction enables modular
 * reductions of products to be calculated rather rapidly, without recourse
 * to annoying things like division.
 *
 * Before starting, you need to do a little work.  In particular, the
 * following things need to be worked out:
 *
 *   * %$m$%, which is the modulus you'll be working with.  This must be odd,
 *     otherwise the whole thing doesn't work.  You're better off using
 *     Barrett reduction if your modulus might be even.
 *
 *   * %$b$%, the radix of the number system you're in (here, it's
 *     @MPW_MAX + 1@).
 *
 *   * %$m' = -m^{-1} \bmod b$%, a useful number for the reduction step.
 *     (This means that the modulus mustn't be even.  This shouldn't be a
 *     problem.)
 *
 *   * %$R = b^n > m > b^{n - 1}$%, or at least %$\log_2 R$%.
 *
 *   * %$R \bmod m$% and %$R^2 \bmod m$%, which are useful when doing
 *     calculations such as exponentiation.
 *
 * Suppose that %$0 \le a_i \le (b^n + b^i - 1) m$% with %$a_i \equiv {}$%
 * %$0 \pmod{b^i}$%.  Let %$w_i = m' a_i/b^i \bmod b$%, and set %$a_{i+1} =
 * a_i + b^i w_i m$%.  Then obviously %$a_{i+1} \equiv {} $% %$a_i
 * \pmod{m}$%, and less obviously %$a_{i+1}/b^i \equiv a_i/b^i + {}$% %$m m'
 * a_i/b^i \equiv 0 \pmod{b}$% so %$a_{i+1} \equiv 0 \pmod{b^{i+1}}$%.
 * Finally, we can bound %$a_{i+1} \le {}$% %$a_i + b^i (b - 1) m = {}$%
 * %$a_i + (b^{i+1} - b^i) m \le (b^n + b^{i+1} - 1) m$%.  As a result, if
 * we're given some %a_0%, we can calculate %$a_n \equiv 0 \pmod{R}$%, with
 * $%a_n \equiv a_0 \pmod{n}$%, i.e., %$a_n/R \equiv a_0 R^{-1} \pmod{m}$%;
 * furthermore, if %$0 \le a_0 < m + b^n%$ then %$0 \le a_n/R < 2 m$%, so a
 * fully reduced result can be obtained with a single conditional
 * subtraction.
 *
 * The result of reduing %$a$% is then %$a R^{-1}$% \bmod m$%.  This is
 * actually rather useful for reducing products, if we run an extra factor of
 * %$R$% through the calculation: the result of reducing the product of
 * %$(x R)(y R) = x y R^2$% is then %$x y R \bmod m$%, preserving the running
 * factor.  Thanks to distributivity, additions and subtractions can be
 * performed on numbers in this form -- the extra factor of %$R$% just runs
 * through all the calculations until it's finally stripped out by a final
 * reduction operation.
 */

/*----- Data structures ---------------------------------------------------*/

/* --- A Montgomery reduction context --- */

typedef struct mpmont {
  mp *m;				/* Modulus */
  mp *mi;				/* %$-m^{-1} \bmod R$% */
  size_t n;				/* %$\log_b R$% */
  mp *r, *r2;				/* %$R \bmod m$%, %$R^2 \bmod m$% */
} mpmont;

/*----- Functions provided ------------------------------------------------*/

/* --- @mpmont_create@ --- *
 *
 * Arguments:	@mpmont *mm@ = pointer to Montgomery reduction context
 *		@mp *m@ = modulus to use
 *
 * Returns:	Zero on success, nonzero on error.
 *
 * Use:		Initializes a Montgomery reduction context ready for use.
 *		The argument @m@ must be a positive odd integer.
 */

extern int mpmont_create(mpmont */*mm*/, mp */*m*/);

/* --- @mpmont_destroy@ --- *
 *
 * Arguments:	@mpmont *mm@ = pointer to a Montgomery reduction context
 *
 * Returns:	---
 *
 * Use:		Disposes of a context when it's no longer of any use to
 *		anyone.
 */

extern void mpmont_destroy(mpmont */*mm*/);

/* --- @mpmont_reduce@ --- *
 *
 * Arguments:	@const mpmont *mm@ = pointer to Montgomery reduction context
 *		@mp *d@ = destination
 *		@mp *a@ = source, assumed positive
 *
 * Returns:	Result, %$a R^{-1} \bmod m$%.
 */

extern mp *mpmont_reduce(const mpmont */*mm*/, mp */*d*/, mp */*a*/);

/* --- @mpmont_mul@ --- *
 *
 * Arguments:	@const mpmont *mm@ = pointer to Montgomery reduction context
 *		@mp *d@ = destination
 *		@mp *a, *b@ = sources, assumed positive
 *
 * Returns:	Result, %$a b R^{-1} \bmod m$%.
 */

extern mp *mpmont_mul(const mpmont */*mm*/, mp */*d*/, mp */*a*/, mp */*b*/);

/* --- @mpmont_expr@ --- *
 *
 * Arguments:	@const mpmont *mm@ = pointer to Montgomery reduction context
 *		@mp *d@ = fake destination
 *		@mp *a@ = base
 *		@mp *e@ = exponent
 *
 * Returns:	Result, %$(a R^{-1})^e R \bmod m$%.  This is useful if
 *		further modular arithmetic is to be performed on the result.
 */

extern mp *mpmont_expr(const mpmont */*mm*/, mp */*d*/,
		       mp */*a*/, mp */*e*/);

/* --- @mpmont_exp@ --- *
 *
 * Arguments:	@const mpmont *mm@ = pointer to Montgomery reduction context
 *		@mp *d@ = fake destination
 *		@mp *a@ = base
 *		@mp *e@ = exponent
 *
 * Returns:	Result, %$a^e \bmod m$%.
 */

extern mp *mpmont_exp(const mpmont */*mm*/, mp */*d*/, mp */*a*/, mp */*e*/);

/* --- @mpmont_mexpr@ --- *
 *
 * Arguments:	@const mpmont *mm@ = pointer to Montgomery reduction context
 *		@mp *d@ = fake destination
 *		@const mp_expfactor *f@ = pointer to array of factors
 *		@size_t n@ = number of factors supplied
 *
 * Returns:	If the bases are %$g_0, g_1, \ldots, g_{n-1}$% and the
 *		exponents are %$e_0, e_1, \ldots, e_{n-1}$% then the result
 *		is:
 *
 *		%$g_0^{e_0} g_1^{e_1} \ldots g_{n-1}^{e_{n-1}} \bmod m$%
 *
 *
 *		except that the %$g_i$% and result are in Montgomery form.
 */

extern mp *mpmont_mexpr(const mpmont */*mm*/, mp */*d*/,
			const mp_expfactor */*f*/, size_t /*n*/);

/* --- @mpmont_mexp@ --- *
 *
 * Arguments:	@const mpmont *mm@ = pointer to Montgomery reduction context
 *		@mp *d@ = fake destination
 *		@const mp_expfactor *f@ = pointer to array of factors
 *		@size_t n@ = number of factors supplied
 *
 * Returns:	Product of bases raised to exponents, all mod @m@.
 *
 * Use:		Convenient interface over @mpmont_mexpr@.
 */

extern mp *mpmont_mexp(const mpmont */*mm*/, mp */*d*/,
		       const mp_expfactor */*f*/, size_t /*n*/);

/*----- That's all, folks -------------------------------------------------*/

#ifdef __cplusplus
  }
#endif

#endif
