/* -*-c-*-
 *
 * Definitions for output feedback mode
 *
 * (c) 1999 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

#ifndef CATACOMB_OFB_DEF_H
#define CATACOMB_OFB_DEF_H

#ifdef __cplusplus
  extern "C" {
#endif

/*----- Header files ------------------------------------------------------*/

#include <stdarg.h>
#include <string.h>

#include <mLib/bits.h>
#include <mLib/sub.h>

#ifndef CATACOMB_ARENA_H
#  include "arena.h"
#endif

#ifndef CATACOMB_BLKC_H
#  include "blkc.h"
#endif

#ifndef CATACOMB_GCIPHER_H
#  include "gcipher.h"
#endif

#ifndef CATACOMB_PARANOIA_H
#  include "paranoia.h"
#endif

#ifndef CATACOMB_RSVR_H
#  include "rsvr.h"
#endif

/*----- Macros ------------------------------------------------------------*/

/* --- @OFB_DEF@ --- *
 *
 * Arguments:	@PRE@, @pre@ = prefixes for the underlying block cipher
 *
 * Use:		Creates definitions for output feedback mode.
 */

#define OFB_DEF(PRE, pre) OFB_DEFX(PRE, pre, #pre, #pre)

#define OFB_DEFX(PRE, pre, name, fname)					\
									\
/* --- @pre_ofbgetiv@ --- *						\
 *									\
 * Arguments:	@const pre_ofbctx *ctx@ = pointer to OFB context block	\
 *		@void *iv@ = pointer to output data block		\
 *									\
 * Returns:	---							\
 *									\
 * Use:		Reads the currently set IV.  Reading and setting an IV	\
 *		is not transparent to the cipher.  It will add a `step'	\
 *		which must be matched by a similar operation during	\
 *		decryption.						\
 */									\
									\
void pre##_ofbgetiv(const pre##_ofbctx *ctx, void *iv)			\
{									\
  octet *p = iv;							\
  unsigned off = ctx->off;						\
  unsigned rest = PRE##_BLKSZ - off;					\
									\
  memcpy(p, ctx->b + off, rest);					\
  memcpy(p + rest, ctx->b, off);					\
}									\
									\
/* --- @pre_ofbsetiv@ --- *						\
 *									\
 * Arguments:	@pre_ofbctx *ctx@ = pointer to OFB context block	\
 *		@cnost void *iv@ = pointer to IV to set			\
 *									\
 * Returns:	---							\
 *									\
 * Use:		Sets the IV to use for subsequent encryption.		\
 */									\
									\
void pre##_ofbsetiv(pre##_ofbctx *ctx, const void *iv)			\
  { memcpy(ctx->b, iv, PRE##_BLKSZ); ctx->off = 0; }			\
									\
/* --- @pre_ofbbdry@ --- *						\
 *									\
 * Arguments:	@pre_ofbctx *ctx@ = pointer to OFB context block	\
 *									\
 * Returns:	---							\
 *									\
 * Use:		Inserts a boundary during encryption.  Successful	\
 *		decryption must place a similar boundary.		\
 */									\
									\
void pre##_ofbbdry(pre##_ofbctx *ctx)					\
{									\
  uint32 t[PRE##_BLKSZ/4];						\
									\
  BLKC_LOAD(PRE, t, ctx->b);						\
  pre##_eblk(&ctx->ctx, t, t);						\
  BLKC_STORE(PRE, ctx->b, t);						\
  ctx->off = 0;								\
  BURN(t);								\
}									\
									\
/* --- @pre_ofbsetkey@ --- *						\
 *									\
 * Arguments:	@pre_ofbctx *ctx@ = pointer to OFB context block	\
 *		@const pre_ctx *k@ = pointer to cipher context		\
 *									\
 * Returns:	---							\
 *									\
 * Use:		Sets the OFB context to use a different cipher key.	\
 */									\
									\
void pre##_ofbsetkey(pre##_ofbctx *ctx, const pre##_ctx *k)		\
  { ctx->ctx = *k; }							\
									\
/* --- @pre_ofbinit@ --- *						\
 *									\
 * Arguments:	@pre_ofbctx *ctx@ = pointer to cipher context		\
 *		@const void *key@ = pointer to the key buffer		\
 *		@size_t sz@ = size of the key				\
 *		@const void *iv@ = pointer to initialization vector	\
 *									\
 * Returns:	---							\
 *									\
 * Use:		Initializes a OFB context ready for use.  You should	\
 *		ensure that the IV chosen is unique: reusing an IV will	\
 *		compromise the security of the entire plaintext.  This	\
 *		is equivalent to calls to @pre_init@, @pre_ofbsetkey@	\
 *		and @pre_ofbsetiv@.					\
 */									\
									\
void pre##_ofbinit(pre##_ofbctx *ctx,					\
		     const void *key, size_t sz,			\
		     const void *iv)					\
{									\
  static const octet zero[PRE##_BLKSZ] = { 0 };				\
									\
  pre##_init(&ctx->ctx, key, sz);					\
  pre##_ofbsetiv(ctx, iv ? iv : zero);					\
}									\
									\
/* --- @pre_ofbencrypt@ --- *						\
 *									\
 * Arguments:	@pre_ofbctx *ctx@ = pointer to OFB context block	\
 *		@const void *src@ = pointer to source data		\
 *		@void *dest@ = pointer to destination data		\
 *		@size_t sz@ = size of block to be encrypted		\
 *									\
 * Returns:	---							\
 *									\
 * Use:		Encrypts or decrypts a block with a block cipher in OFB	\
 *		mode: encryption and decryption are the same in OFB.	\
 *		The destination may be null to just churn the feedback	\
 *		round for a bit.  The source may be null to use the	\
 *		cipher as a random data generator.			\
 */									\
									\
static const rsvr_policy pre##_ofbpolicy = { 0, PRE##_BLKSZ, PRE##_BLKSZ }; \
									\
void pre##_ofbencrypt(pre##_ofbctx *ctx,				\
		      const void *src, void *dest,			\
		      size_t sz)					\
{									\
  rsvr_plan plan;							\
  const octet *s = src, *p;						\
  octet *d = dest;							\
  uint32 t[PRE##_BLKSZ/4], u[PRE##_BLKSZ/4];				\
									\
  /* Construct a plan and prepare to follow through. */			\
  rsvr_mkplan(&plan, &pre##_ofbpolicy, ctx->off, sz);			\
  BLKC_LOAD(PRE, t, ctx->b);						\
									\
  /* Initial portion, fulfilled from the buffer.  If the chunk is small	\
   * enough, then this will be the only portion.  If the buffer is	\
   * currently empty, then we must prepare it.				\
   */									\
  if (plan.head) {							\
    if (!ctx->off) {							\
      pre##_eblk(&ctx->ctx, t, t);					\
      BLKC_STORE(PRE, ctx->b, t);					\
    }									\
    p = ctx->b + ctx->off; ctx->off += plan.head;			\
    if (!d) /* nothing to do */;					\
    else if (!s) { memcpy(d, p, plan.head); d += plan.head; }		\
    else while (plan.head--) *d++ = *s++ ^ *p++;			\
  }									\
									\
  /* If the buffer is all used, then reset it ready for next time. */	\
  ctx->off -= plan.from_rsvr;						\
									\
  /* Handle multiple whole blocks. */					\
  if (!d) while (plan.from_input) {					\
    pre##_eblk(&ctx->ctx, t, t);					\
    plan.from_input -= PRE##_BLKSZ;					\
  } else if (!s) while (plan.from_input) {				\
    pre##_eblk(&ctx->ctx, t, t);					\
    BLKC_STORE(PRE, d, t); d += PRE##_BLKSZ;				\
    plan.from_input -= PRE##_BLKSZ;					\
  } else while (plan.from_input) {					\
    pre##_eblk(&ctx->ctx, t, t);					\
    BLKC_LOAD(PRE, u, s); s += PRE##_BLKSZ;				\
    BLKC_XSTORE(PRE, d, t, u); d += PRE##_BLKSZ;			\
    plan.from_input -= PRE##_BLKSZ;					\
  }									\
									\
  /* Final portion.  Note that the buffer must be empty if there is a	\
   * tail, since otherwise the input data would have been part of the	\
   * head portion instad. */						\
  if (!plan.tail)							\
    BLKC_STORE(PRE, ctx->b, t);						\
  else {								\
    pre##_eblk(&ctx->ctx, t, t);					\
    BLKC_STORE(PRE, ctx->b, t);						\
    p = ctx->b; ctx->off += plan.tail;					\
    if (!d) /* nothing to do */;					\
    else if (!s) { memcpy(d, p, plan.tail); d += plan.tail; }		\
    else while (plan.tail--) *d++ = *s++ ^ *p++;			\
  }									\
}									\
									\
/* --- Generic cipher interface --- */					\
									\
static const gcipher_ops gops;						\
									\
typedef struct gctx {							\
  gcipher c;								\
  pre##_ofbctx k;							\
} gctx;									\
									\
static gcipher *ginit(const void *k, size_t sz)				\
{									\
  gctx *g = S_CREATE(gctx);						\
  g->c.ops = &gops;							\
  pre##_ofbinit(&g->k, k, sz, 0);					\
  return (&g->c);							\
}									\
									\
static void gencrypt(gcipher *c, const void *s, void *t, size_t sz)	\
  { gctx *g = (gctx *)c; pre##_ofbencrypt(&g->k, s, t, sz); }		\
									\
static void gdestroy(gcipher *c)					\
  { gctx *g = (gctx *)c; BURN(*g); S_DESTROY(g); }			\
									\
static void gsetiv(gcipher *c, const void *iv)				\
  { gctx *g = (gctx *)c; pre##_ofbsetiv(&g->k, iv); }			\
									\
static void gbdry(gcipher *c)						\
{									\
  gctx *g = (gctx *)c;							\
  pre##_ofbbdry(&g->k);							\
}									\
									\
static const gcipher_ops gops = {					\
  &pre##_ofb,								\
  gencrypt, gencrypt, gdestroy, gsetiv, gbdry				\
};									\
									\
const gccipher pre##_ofb = {						\
  name "-ofb", pre##_keysz, PRE##_BLKSZ,				\
  ginit									\
};									\
									\
/* --- Generic random number generator interface --- */			\
									\
typedef struct grctx {							\
  grand r;								\
  pre##_ofbctx k;							\
} grctx;								\
									\
static void grdestroy(grand *r)						\
  { grctx *g = (grctx *)r; BURN(*g); S_DESTROY(g); }			\
									\
static int grmisc(grand *r, unsigned op, ...)				\
{									\
  grctx *g = (grctx *)r;						\
  va_list ap;								\
  int rc = 0;								\
  uint32 i;								\
  octet buf[PRE##_BLKSZ];						\
  va_start(ap, op);							\
									\
  switch (op) {								\
    case GRAND_CHECK:							\
      switch (va_arg(ap, unsigned)) {					\
	case GRAND_CHECK:						\
	case GRAND_SEEDINT:						\
	case GRAND_SEEDUINT32:						\
	case GRAND_SEEDBLOCK:						\
	case GRAND_SEEDRAND:						\
	  rc = 1;							\
	  break;							\
	default:							\
	  rc = 0;							\
	  break;							\
      }									\
      break;								\
    case GRAND_SEEDINT:							\
      memset(buf, 0, sizeof(buf));					\
      i = va_arg(ap, unsigned);						\
      STORE32(buf, i);							\
      pre##_ofbsetiv(&g->k, buf);					\
      break;								\
    case GRAND_SEEDUINT32:						\
      memset(buf, 0, sizeof(buf));					\
      i = va_arg(ap, uint32);						\
      STORE32(buf, i);							\
      pre##_ofbsetiv(&g->k, buf);					\
      break;								\
    case GRAND_SEEDBLOCK: {						\
      const void *p = va_arg(ap, const void *);				\
      size_t sz = va_arg(ap, size_t);					\
      if (sz < sizeof(buf)) {						\
	memset(buf, 0, sizeof(buf));					\
	memcpy(buf, p, sz);						\
	p = buf;							\
      }									\
      pre##_ofbsetiv(&g->k, p);						\
    } break;								\
    case GRAND_SEEDRAND: {						\
      grand *rr = va_arg(ap, grand *);					\
      rr->ops->fill(rr, buf, sizeof(buf));				\
      pre##_ofbsetiv(&g->k, buf);					\
    } break;								\
    default:								\
      GRAND_BADOP;							\
      break;								\
  }									\
									\
  va_end(ap);								\
  return (rc);								\
}									\
									\
static octet grbyte(grand *r)						\
{									\
  grctx *g = (grctx *)r;						\
  octet o;								\
  pre##_ofbencrypt(&g->k, 0, &o, 1);					\
  return (o);								\
}									\
									\
static uint32 grword(grand *r)						\
{									\
  grctx *g = (grctx *)r;						\
  octet b[4];								\
  pre##_ofbencrypt(&g->k, 0, b, sizeof(b));				\
  return (LOAD32(b));							\
}									\
									\
static void grfill(grand *r, void *p, size_t sz)			\
  { grctx *g = (grctx *)r; pre##_ofbencrypt(&g->k, 0, p, sz); }		\
									\
static const grand_ops grops = {					\
  name "-ofb",								\
  GRAND_CRYPTO, 0,							\
  grmisc, grdestroy,							\
  grword, grbyte, grword, grand_defaultrange, grfill			\
};									\
									\
/* --- @pre_ofbrand@ --- *						\
 *									\
 * Arguments:	@const void *k@ = pointer to key material		\
 *		@size_t sz@ = size of key material			\
 *									\
 * Returns:	Pointer to generic random number generator interface.	\
 *									\
 * Use:		Creates a random number interface wrapper around an	\
 *		OFB-mode block cipher.					\
 */									\
									\
grand *pre##_ofbrand(const void *k, size_t sz)				\
{									\
  grctx *g = S_CREATE(grctx);						\
  g->r.ops = &grops;							\
  pre##_ofbinit(&g->k, k, sz, 0);					\
  return (&g->r);							\
}									\
									\
OFB_TESTX(PRE, pre, name, name)

/*----- Test rig ----------------------------------------------------------*/

#define OFB_TEST(PRE, pre) OFB_TESTX(PRE, pre, #pre, #pre)

#ifdef TEST_RIG

#include "modes-test.h"

/* --- @OFB_TEST@ --- *
 *
 * Arguments:	@PRE@, @pre@ = prefixes for block cipher definitions
 *
 * Use:		Standard test rig for OFB functions.
 */

#define OFB_TESTX(PRE, pre, name, fname)				\
									\
static pre##_ctx key;							\
static pre##_ofbctx ctx;						\
									\
static void pre##_ofb_test_setup(const octet *k, size_t ksz)		\
  { pre##_init(&key, k, ksz); pre##_ofbsetkey(&ctx, &key); }		\
									\
static void pre##_ofb_test_reset(const octet *iv)			\
  { pre##_ofbsetiv(&ctx, iv); }						\
									\
static void pre##_ofb_test_enc(const octet *s, octet *d, size_t sz)	\
  { pre##_ofbencrypt(&ctx, s, d, sz); }					\
									\
int main(int argc, char *argv[])					\
{									\
  return test_encmode(fname "-ofb", PRE##_KEYSZ, PRE##_BLKSZ, 1, 0,	\
		      pre##_ofb_test_setup, pre##_ofb_test_reset,	\
		      pre##_ofb_test_enc, pre##_ofb_test_enc,		\
		      argc, argv);					\
}

#else
#  define OFB_TESTX(PRE, pre, name, fname)
#endif

/*----- That's all, folks -------------------------------------------------*/

#ifdef __cplusplus
  }
#endif

#endif
