/* -*-c-*-
 *
 * The RIPEMD-128 message digest function
 *
 * (c) 1998 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Notes on the RIPEMD-128 hash function -----------------------------*
 *
 * RIPEMD-128 was invented by Hans Dobbertin, Antoon Bosselaers and Bart
 * Preneel, as a drop-in replacement for MD5 (with the same sized output).
 * It's a cut-down version of RIPEMD-160, which should be used in preference.
 */

#ifndef CATACOMB_RMD128_H
#define CATACOMB_RMD128_H

#ifdef __cplusplus
  extern "C" {
#endif

/*----- Header files ------------------------------------------------------*/

#include <mLib/bits.h>

#ifndef CATACOMB_GHASH_H
#  include "ghash.h"
#endif

/*----- Magic numbers -----------------------------------------------------*/

#define RMD128_BUFSZ 64
#define RMD128_HASHSZ 16
#define RMD128_STATESZ 16

/*----- Data structures ---------------------------------------------------*/

typedef struct rmd128_ctx {
  uint32 a, b, c, d;			/* Chaining variables */
  uint32 nl, nh;			/* Byte count so far */
  unsigned off;				/* Offset into buffer */
  octet buf[RMD128_BUFSZ];		/* Accumulation buffer */
} rmd128_ctx;

/*----- Functions provided ------------------------------------------------*/

/* --- @rmd128_compress@ --- *
 *
 * Arguments:	@rmd128_ctx *ctx@ = pointer to context block
 *		@const void *sbuf@ = pointer to buffer of appropriate size
 *
 * Returns:	---
 *
 * Use:		RIPEMD-128 compression function.
 */

extern void rmd128_compress(rmd128_ctx */*ctx*/, const void */*sbuf*/);

/* --- @rmd128_init@ --- *
 *
 * Arguments:	@rmd128_ctx *ctx@ = pointer to context block to initialize
 *
 * Returns:	---
 *
 * Use:		Initializes a context block ready for hashing.
 */

extern void rmd128_init(rmd128_ctx */*ctx*/);

/* --- @rmd128_set@ --- *
 *
 * Arguments:	@rmd128_ctx *ctx@ = pointer to context block
 *		@const void *buf@ = pointer to state buffer
 *		@unsigned long count@ = current count of bytes processed
 *
 * Returns:	---
 *
 * Use:		Initializes a context block from a given state.  This is
 *		useful in cases where the initial hash state is meant to be
 *		secret, e.g., for NMAC and HMAC support.
 */

extern void rmd128_set(rmd128_ctx */*ctx*/,
		       const void */*buf*/, unsigned long /*count*/);

/* --- @rmd128_hash@ --- *
 *
 * Arguments:	@rmd128_ctx *ctx@ = pointer to context block
 *		@const void *buf@ = buffer of data to hash
 *		@size_t sz@ = size of buffer to hash
 *
 * Returns:	---
 *
 * Use:		Hashes a buffer of data.  The buffer may be of any size and
 *		alignment.
 */

extern void rmd128_hash(rmd128_ctx */*ctx*/,
			const void */*buf*/, size_t /*sz*/);

/* --- @rmd128_done@ --- *
 *
 * Arguments:	@rmd128_ctx *ctx@ = pointer to context block
 *		@void *hash@ = pointer to output buffer
 *
 * Returns:	---
 *
 * Use:		Returns the hash of the data read so far.
 */

extern void rmd128_done(rmd128_ctx */*ctx*/, void */*hash*/);

/* --- @rmd128_state@ --- *
 *
 * Arguments:	@rmd128_ctx *ctx@ = pointer to context
 *		@void *state@ = pointer to buffer for current state
 *
 * Returns:	Number of bytes written to the hash function so far.
 *
 * Use:		Returns the current state of the hash function such that
 *		it can be passed to @rmd128_set@.
 */

extern unsigned long rmd128_state(rmd128_ctx */*ctx*/, void */*state*/);

/*----- Generic hash interface --------------------------------------------*/

extern const gchash rmd128;

/*----- That's all, folks -------------------------------------------------*/

#ifdef __cplusplus
  }
#endif

#endif
