/* -*-c-*-
 *
 * Salsa20 core definitions
 *
 * (c) 2015 Straylight/Edgeware
 */

#ifndef CATACOMB_SALSA20_CORE_H
#define CATACOMB_SALSA20_CORE_H

#ifdef __cplusplus
  extern "C" {
#endif

/*----- Header files ------------------------------------------------------*/

#include <mLib/bits.h>
#include <mLib/macros.h>

#ifndef CATACOMB_SALSA20_H
#  include "salsa20.h"
#endif

/*----- Magic constants ---------------------------------------------------*/

/* The magic Salsa20 constants, for 256-bit keys... */
#define SALSA20_A256 0x61707865		/* e x p a */
#define SALSA20_B256 0x3320646e		/* n d   3 */
#define SALSA20_C256 0x79622d32		/* 2 - b y */
#define SALSA20_D256 0x6b206574		/* t e   k */

/* ... and for 128-bit keys ... */
#define SALSA20_A128 SALSA20_A256	/* e x p a */
#define SALSA20_B128 0x3120646e		/* n d   1 */
#define SALSA20_C128 0x79622d36		/* 6 - b y */
#define SALSA20_D128 SALSA20_D256	/* t e   k */

/* ... and for 80-bit keys, for completeness's sake. */
#define SALSA20_A80 SALSA20_A128	/* e x p a */
#define SALSA20_B80 SALSA20_B128	/* n d   1 */
#define SALSA20_C80 0x79622d30		/* 0 - b y */
#define SALSA20_D80 SALSA20_D128	/* t e   k */

/*----- The Salsa20 core function -----------------------------------------*/

/* It makes life somewhat easier if we don't actually store and maintain the
 * input matrix in the textbook order.  Instead, we rotate the columns other
 * than the leftmost one upwards, so that the constants which were originally
 * along the diagonal end up on the top row.  We'll need to undo this
 * permutation on output, but that's not too terrible an imposition.
 *
 * The permutation we're applying to the matrix elements is this:
 *
 * [  0  1  2  3 ]	 [  0  5 10 15 ]
 * [  4  5  6  7 ]  -->  [  4  9 14  3 ]
 * [  8  9 10 11 ]	 [  8 13  2  7 ]
 * [ 12 13 14 15 ]	 [ 12  1  6 11 ]
 *
 * and as a result, we need to apply this inverse permutation to figure out
 * which indices to use in the doublerow function and elsewhere.
 *
 * [  0 13 10  7 ]
 * [  4  1 14 11 ]
 * [  8  5  2 15 ]
 * [ 12  9  6  3 ]
 */

/* The Salsa20 quarter-round.  Read from the matrix @y@ at indices @a@, @b@,
 * @c@, and @d@; and write to the corresponding elements of @z@.
 */
#define SALSA20_QR(z, y, a, b, c, d) do {				\
  (z)[b] = (y)[b] ^ ROL32((y)[a] + (y)[d],  7);				\
  (z)[c] = (y)[c] ^ ROL32((z)[b] + (y)[a],  9);				\
  (z)[d] = (y)[d] ^ ROL32((z)[c] + (z)[b], 13);				\
  (z)[a] = (y)[a] ^ ROL32((z)[d] + (z)[c], 18);				\
} while (0)

/* The Salsa20 double-round.  Read from matrix @y@, writing the result to
 * @z@.
 */
#define SALSA20_DR(z, y) do {						\
  SALSA20_QR(z, y,  0,  4,  8, 12);					\
  SALSA20_QR(z, y,  1,  5,  9, 13);					\
  SALSA20_QR(z, y,  2,  6, 10, 14);					\
  SALSA20_QR(z, y,  3,  7, 11, 15);					\
  SALSA20_QR(z, z,  0, 13, 10,  7);					\
  SALSA20_QR(z, z,  1, 14, 11,  4);					\
  SALSA20_QR(z, z,  2, 15,  8,  5);					\
  SALSA20_QR(z, z,  3, 12,  9,  6);					\
} while (0)

/* The Salsa20 feedforward step, used at the end of the core function.  Here,
 * @y@ contains the original input matrix; @z@ contains the final one, and is
 * updated.  The output is rendered in canonical order, ready for output.
 */
#define SALSA20_FFWD(z, y) do {						\
  const uint32 *_y = (y);						\
  uint32 *_z = (z);							\
  int _t;								\
  _z[ 0] = _z[ 0] + _y[ 0]; _z[ 4] = _z[ 4] + _y[ 4];			\
  _z[ 8] = _z[ 8] + _y[ 8]; _z[12] = _z[12] + _y[12];			\
      _t = _z[ 1] + _y[ 1]; _z[ 1] = _z[13] + _y[13];			\
  _z[13] = _z[ 9] + _y[ 9]; _z[ 9] = _z[ 5] + _y[ 5]; _z[ 5] = _t;	\
      _t = _z[ 2] + _y[ 2]; _z[ 2] = _z[10] + _y[10]; _z[10] = _t;	\
      _t = _z[ 6] + _y[ 6]; _z[ 6] = _z[14] + _y[14]; _z[14] = _t;	\
      _t = _z[ 3] + _y[ 3]; _z[ 3] = _z[ 7] + _y[ 7];			\
  _z[ 7] = _z[11] + _y[11]; _z[11] = _z[15] + _y[15]; _z[15] = _t;	\
} while (0)

/* Various numbers of rounds, unrolled.  Read from @y@, and write to @z@. */
#define SALSA20_4R(z, y)						\
  do { SALSA20_DR(z, y); SALSA20_DR(z, z); } while (0)
#define SALSA20_8R(z, y)						\
  do { SALSA20_4R(z, y); SALSA20_4R(z, z); } while (0)
#define SALSA20_12R(z, y)						\
  do { SALSA20_8R(z, y); SALSA20_4R(z, z); } while (0)
#define SALSA20_20R(z, y)						\
  do { SALSA20_12R(z, y); SALSA20_8R(z, z); } while (0)

/* Apply @n@ (must be even) rounds, rolled.  (This seems to be faster,
 * probably because it fits in cache better).  Read from @y@, and write to
 * @z@.
 */
#define SALSA20_nR(z, y, n) do {					\
  int _i;								\
  SALSA20_DR(z, y);							\
  for (_i = 0; _i < (n)/2 - 1; _i++) SALSA20_DR(z, z);			\
} while (0)

/* Step the counter in the Salsa20 state matrix @a@. */
#define SALSA20_STEP(a)							\
  do { (a)[8] = U32((a)[8] + 1); (a)[5] += !(a)[8]; } while (0)

/*----- Buffering and output ----------------------------------------------*
 *
 * These macros are also used by ChaCha.
 */

/* Copy the Salsa20 matrix @a@ to the output buffer at @d@, advancing @d@
 * past the new material.
 */
#define SALSA20_GENFULL(a, d) do {					\
  int _i;								\
									\
  for (_i = 0; _i < 16; _i++) { STORE32_L((d), (a)[_i]); (d) += 4; }	\
} while (0)

/* XOR the contents the input buffer at @s@ with the Salsa20 matrix @a@,
 * writing the result to @d@ and advance @s@ and @d@.
 */
#define SALSA20_MIXFULL(a, d, s) do {					\
  uint32 _x;								\
  int _i;								\
									\
  for (_i = 0; _i < 16; _i++) {						\
    _x = LOAD32_L(s); (s) += 4;						\
    _x ^= (a)[_i];							\
    STORE32_L((d), _x); (d) += 4;					\
  }									\
} while (0)

/* Fill the context @ctx@'s buffer from the matrix @a@ in preparation for
 * emitting partial blocks of output.
 */
#define SALSA20_PREPBUF(ctx, a) do {					\
  int _i;								\
  for (_i = 0; _i < 16; _i++) STORE32_L((ctx)->b + 4*_i, (a)[_i]);	\
  (ctx)->off = 0;							\
} while (0)

/* Write at most @n@ bytes of buffered output from the context @ctx@ to the
 * output buffer @d@ (if it's not null), XORing it with the input buffer @s@
 * (if that's not null).  Both @s@ and @d@ are advanced if they aren't null;
 * @n@ is decreased appropriately.
 */
#define SALSA20_OUTBUF(ctx, d, s, n) do {				\
  const octet *_p = (ctx)->b + (ctx)->off;				\
  size_t _n = (n);							\
									\
  (ctx)->off += _n;							\
  if (!(d)) /* nothing to do */;					\
  else if (!(s)) { memcpy((d), _p, _n); (d) += _n; }			\
  else while (_n--) *(d)++ = *(s)++ ^ *_p++;				\
} while (0)

/*----- Variants and naming -----------------------------------------------*/

/* Common numbers of rounds, for which we generate definitions. */
#define SALSA20_VARS(_) _(8) _(12) _(20)

/* Constructing externally-facing names. */
#define SALSA20_DECOR(base, r, suff) SALSA20__DECOR_##r(base, suff)
#define SALSA20__DECOR_20(base, suff) GLUE(base, suff)
#define SALSA20__DECOR_12(base, suff) GLUE(base##12, suff)
#define SALSA20__DECOR_8(base, suff) GLUE(base##8, suff)

/* Preprocessor-time table of the standard names. */
#define SALSA20_NAME_20 "salsa20"
#define SALSA20_NAME_12 "salsa20/12"
#define SALSA20_NAME_8 "salsa20/8"

/*----- That's all, folks -------------------------------------------------*/

#ifdef __cplusplus
  }
#endif

#endif
