/* -*-c-*-
 *
 * The X448 key-agreement algorithm
 *
 * (c) 2017 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

#ifndef CATACOMB_X448_H
#define CATACOMB_X448_H

#ifdef __cplusplus
  extern "C" {
#endif

/*----- Notes on the X448 key-agreement algorithm -------------------------*
 *
 * This is X448, as described in RFC7748, based on the elliptic curve defined
 * in Mike Hamburg, `Ed448-Goldilocks, a new elliptic curve', EUROCRYPT 2016,
 * https://eprint.iacr.org/2015/625/.  (The curve here is 4-isogenous to
 * Hamburg's curve.)
 *
 * The RFC-specified operation is simpler than the Diffie--Hellman function
 * described in Hamburg's paper, since it doesn't involve the `Decaf'
 * cofactor elimination procedure.  Indeed, it looks very much like X25519
 * with Hamburg's curve slotted in in place of Bernstein's.
 */

/*----- Header files ------------------------------------------------------*/

#include <mLib/bits.h>

#ifndef CATACOMB_KEY_H
#  include "key.h"
#endif

/*----- Key fetching ------------------------------------------------------*/

typedef struct x448_priv { key_bin priv, pub; } x448_priv;
typedef struct x448_pub { key_bin pub; } x448_pub;

extern const key_fetchdef x448_pubfetch[], x448_privfetch[];
#define X448_PUBFETCHSZ 3
#define X448_PRIVFETCHSZ 6

/*----- Important constants -----------------------------------------------*/

#define X448_KEYSZ 56
#define X448_PUBSZ 56
#define X448_OUTSZ 56

extern const octet x448_base[56];

/*----- Functions provided ------------------------------------------------*/

/* --- @x448@ --- *
 *
 * Arguments:	@octet zz[X448_OUTSZ]@ = where to put the result
 *		@const octet k[X448_KEYSZ]@ = pointer to private key
 *		@const octet qx[X448_PUBSZ]@ = pointer to public value
 *
 * Returns:	---
 *
 * Use:		Calculates X448 of @k@ and @qx@.
 */

extern void x448(octet /*zz*/[X448_OUTSZ],
		 const octet /*k*/[X448_KEYSZ],
		 const octet /*qx*/[X448_PUBSZ]);

/*----- That's all, folks -------------------------------------------------*/

#ifdef __cplusplus
  }
#endif

#endif
